#include <stdint.h>
#include <stdbool.h>


#define PWRCON(i)     (*((volatile uint32_t*)(0x3c500000 \
                                            + ((i) == 4 ? 0x6c : \
                                              ((i) == 3 ? 0x68 : \
                                              ((i) == 2 ? 0x58 : \
                                              ((i) == 1 ? 0x4c : \
                                                          0x48)))))))
#define PCON(i)       (*((volatile uint32_t*)(0x3cf00000 + ((i) << 5))))
#define PDAT(i)       (*((volatile uint32_t*)(0x3cf00004 + ((i) << 5))))
#define GPIOCMD       (*((volatile uint32_t*)(0x3cf00200)))
#define SPIBASE(i)      ((i) == 2 ? 0x3d200000 : \
                         (i) == 1 ? 0x3ce00000 : \
                                    0x3c300000)
#define SPICLKGATE(i)   ((i) == 2 ? 0x2f : \
                         (i) == 1 ? 0x2b : \
                                    0x22)
#define SPICTRL(i)    (*((volatile uint32_t*)(SPIBASE(i))))
#define SPISETUP(i)   (*((volatile uint32_t*)(SPIBASE(i) + 0x4)))
#define SPISTATUS(i)  (*((volatile uint32_t*)(SPIBASE(i) + 0x8)))
#define SPIUNKREG1(i) (*((volatile uint32_t*)(SPIBASE(i) + 0xc)))
#define SPITXDATA(i)  (*((volatile uint32_t*)(SPIBASE(i) + 0x10)))
#define SPIRXDATA(i)  (*((volatile uint32_t*)(SPIBASE(i) + 0x20)))
#define SPICLKDIV(i)  (*((volatile uint32_t*)(SPIBASE(i) + 0x30)))
#define SPIRXLIMIT(i) (*((volatile uint32_t*)(SPIBASE(i) + 0x34)))
#define SPIUNKREG3(i) (*((volatile uint32_t*)(SPIBASE(i) + 0x38)))
#define AESCONTROL    (*((volatile uint32_t*)(0x38c00000)))
#define AESGO         (*((volatile uint32_t*)(0x38c00004)))
#define AESUNKREG0    (*((volatile uint32_t*)(0x38c00008)))
#define AESSTATUS     (*((volatile uint32_t*)(0x38c0000c)))
#define AESUNKREG1    (*((volatile uint32_t*)(0x38c00010)))
#define AESKEYLEN     (*((volatile uint32_t*)(0x38c00014)))
#define AESOUTSIZE    (*((volatile uint32_t*)(0x38c00018)))
#define AESOUTADDR    (*((volatile uint32_t*)(0x38c00020)))
#define AESINSIZE     (*((volatile uint32_t*)(0x38c00024)))
#define AESINADDR     (*((volatile uint32_t*)(0x38c00028)))
#define AESAUXSIZE    (*((volatile uint32_t*)(0x38c0002c)))
#define AESAUXADDR    (*((volatile uint32_t*)(0x38c00030)))
#define AESSIZE3      (*((volatile uint32_t*)(0x38c00034)))
#define AESKEY          ((volatile uint32_t*)(0x38c0004c))
#define AESTYPE       (*((volatile uint32_t*)(0x38c0006c)))
#define AESIV           ((volatile uint32_t*)(0x38c00074))
#define AESTYPE2      (*((volatile uint32_t*)(0x38c00088)))
#define AESUNKREG2    (*((volatile uint32_t*)(0x38c0008c)))


void clockgate_enable(int gate, bool enable)
{
    if (enable) PWRCON(gate >> 5) &= ~(1 << (gate & 0x1f));
    else PWRCON(gate >> 5) |= 1 << (gate & 0x1f);
}

void spi_prepare(int port)
{
    clockgate_enable(SPICLKGATE(port), true);
    SPISTATUS(port) = 0xf;
    SPICTRL(port) |= 0xc;
    SPICLKDIV(port) = 4;
    SPIUNKREG1(port) = 6;
    SPISETUP(port) = 0x10618;
    SPICTRL(port) |= 0xc;
    SPICTRL(port) = 1;
}

void spi_release(int port)
{
    clockgate_enable(SPICLKGATE(port), false);
}

uint32_t spi_write(int port, uint32_t data)
{
    SPIRXLIMIT(port) = 1;
    while ((SPISTATUS(port) & 0x1f0) == 0x100);
    SPITXDATA(port) = data;
    while (!(SPISTATUS(port) & 0x3e00));
    return SPIRXDATA(port);
}

void spi_read(int port, uint32_t size, void* buf)
{
    uint8_t* buffer = (uint8_t*)buf;
    SPIRXLIMIT(port) = size;
    SPISETUP(port) |= 1;
    while (size--)
    {
        while (!(SPISTATUS(port) & 0x3e00));
        *buffer++ = SPIRXDATA(port);
    }
    SPISETUP(port) &= ~1;
}

void bootflash_ce(int port, bool state)
{
    if (state)
    {
        if (port == 2) PDAT(0xe) &= ~0x40;
        else if (port == 1) PDAT(6) &= ~0x10;
        else PDAT(0) &= ~1;
    }
    else
    {
        if (port == 2) PDAT(0xe) |= 0x40;
        else if (port == 1) PDAT(6) |= 0x10;
        else PDAT(0) |= 1;
    }
}

void bootflash_init()
{
    PCON(0) = (PCON(0) & ~0xffff) | 0x2222;
    GPIOCMD = 0xf;
    PCON(6) = (PCON(6) & ~0xffff0000) | 0x22220000;
    GPIOCMD = 0x6040f;
    PCON(0xe) = (PCON(0xe) & ~0xff000000) | 0x22000000;
    PCON(0xf) = (PCON(0xf) & ~0xff) | 0x22;
    GPIOCMD = 0xe060f;
}

void bootflash_read(int port, uint32_t addr, uint32_t size, void* buf)
{
    spi_prepare(port);
    while (true)
    {
        bootflash_ce(port, true);
        spi_write(port, 5);
        if (!(spi_write(port, 0xff) & 1)) break;
        bootflash_ce(port, false);
    }
    bootflash_ce(port, false);
    bootflash_ce(port, true);
    spi_write(port, 3);
    spi_write(port, (addr >> 16) & 0xff);
    spi_write(port, (addr >> 8) & 0xff);
    spi_write(port, addr & 0xff);
    spi_read(port, size, buf);
    bootflash_ce(port, false);
    spi_release(port);
}

enum hwkeyaes_direction
{
    HWKEYAES_DECRYPT = 0,
    HWKEYAES_ENCRYPT = 1
};

void hwkeyaes(enum hwkeyaes_direction direction, uint32_t keyidx, void* data, uint32_t size)
{
    int i;
    clockgate_enable(7, true);
    for (i = 0; i < 4; i++) AESIV[i] = 0;
    AESUNKREG0 = 1;
    AESUNKREG0 = 0;
    AESCONTROL = 1;
    AESUNKREG1 = 0;
    AESTYPE = keyidx;
    AESTYPE2 = ~AESTYPE;
    AESUNKREG2 = 0;
    AESKEYLEN = direction == HWKEYAES_ENCRYPT ? 9 : 8;
    AESOUTSIZE = size;
    AESOUTADDR = (uint32_t)data;
    AESINSIZE = size;
    AESINADDR = (uint32_t)data;
    AESAUXSIZE = size;
    AESAUXADDR = (uint32_t)data;
    AESSIZE3 = size;
    AESGO = 1;
    while (!(AESSTATUS & 0xf));
    clockgate_enable(7, false);
}

void main(void)
{
    bootflash_init();
    bootflash_read(0, 0x8000, 0x800, (void*)0x22020000);
    bootflash_read(0, 0x8800, *((uint32_t*)0x2202000c), (void*)0x22020800);
    hwkeyaes(HWKEYAES_DECRYPT, 2, (void*)0x22020010, 16);
    hwkeyaes(HWKEYAES_DECRYPT, 2, (void*)0x22020040, 16);
    hwkeyaes(HWKEYAES_DECRYPT, 2, (void*)0x22020800, *((uint32_t*)0x2202000c));
}
